// Copyright 2020 Ammar Yasser. All rights reserved.
// Use of this source code is governed by a BSD 2-Clause "Simplified" license
// that can be found in the LICENSE file.

library linear_structures;

/// Complexity of each operation
/// ----------------------------
///  operation   | complexity
///  peekFirst   |   O(1)
///  peekLast    |   O(1)
///  isEmpty     |   O(1)
///  isNotEmpty  |   O(1)
///  length      |   O(1)
///  getAll      |   O(n)
///  addFirst    |   O(1)
///  addLast     |   O(1)
///  removeFirst |   O(1)
///  removeLast  |   O(1)
///  clear       |   O(1)

/// Implementation of Doubly LinkedList
class DoublyLinkedList<T> {
  int _size = 0;
  _Node<T>? _head;
  _Node<T>? _tail;

  /// Returns the first node data.
  T? get peekFirst => _head?.data;

  /// Returns the last node data.
  T? get peekLast => _tail?.data;

  /// Returns true if there are no nodes in this list.
  bool get isEmpty => _head == null;

  /// Returns true if there is at least one node in this list.
  bool get isNotEmpty => _head != null;

  /// Returns the number of nodes in this list.
  int get length => _size;

  /// Gets all nodes data in this LinkedList and then returns them as type List.
  /// This method returns null if the list is empty.
  List<T?>? getAll() {
    if (_head == null) return null;
    if (_head == _tail) return [_head!.data];
    List<T?> nodes = [];
    _Node<T>? last = _head;
    while (last != null) {
      nodes.add(last.data);
      last = last.next;
    }
    return nodes;
  }

  /// Adds a node to the beginning of the list.
  void addFirst(T obj) {
    _Node<T> newNode = _Node(obj);
    newNode.next = _head;
    newNode.previous = null;
    if (_head != null) _head!.previous = newNode;
    _head = newNode;
    if (_tail == null) _tail = newNode;
    _size++;
  }

  /// Adds a node to the end of the list.
  void addLast(T obj) {
    _Node<T> newNode = _Node(obj);
    newNode.next = null;
    newNode.previous = _tail;
    if (_tail != null) _tail!.next = newNode;
    _tail = newNode;
    if (_head == null) _head = newNode;
    _size++;
  }

  /// Removes and returns the first node data from this list.
  /// This method returns null if the list is empty.
  T? removeFirst() {
    if (isEmpty) return null;
    T? removedData = _head!.data;
    _head = _head!.next;
    _head!.previous = null;
    _size--;
    return removedData;
  }

  /// Removes and returns the last node data from this list.
  /// This method returns null if the list is empty.
  T? removeLast() {
    T? removedData = _tail!.data;
    _tail = _tail!.previous;
    _tail!.next = null;
    _size--;
    return removedData;
  }

  void clear() {
    _size = 0;
    _head = null;
    _tail = null;
  }
}

class _Node<T> {
  T? data;
  _Node<T>? previous;
  _Node<T>? next;

  _Node(T obj) {
    data = obj;
    previous = null;
    next = null;
  }
}
